@@ -16,7 +16,8 @@
  * with this program; If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include <xen/config.h>
+asm(".file \"" __FILE__ "\"");
+
 #include <xen/lib.h>
 #include <xen/init.h>
 #include <xen/mm.h>
@@ -41,10 +42,6 @@
 
 unsigned int __read_mostly m2p_compat_vstart = __HYPERVISOR_COMPAT_VIRT_START;
 
-/* Enough page directories to map into the bottom 1GB. */
-l3_pgentry_t __section(".bss.page_aligned") l3_bootmap[L3_PAGETABLE_ENTRIES];
-l2_pgentry_t __section(".bss.page_aligned") l2_bootmap[L2_PAGETABLE_ENTRIES];
-
 l2_pgentry_t *compat_idle_pg_table_l2;
 
 void *do_page_walk(struct vcpu *v, unsigned long addr)
@@ -1365,8 +1362,9 @@ int memory_add(unsigned long spfn, unsigned long epfn, unsigned int pxm)
 
     if ( !valid_numa_range(spfn << PAGE_SHIFT, epfn << PAGE_SHIFT, node) )
     {
-        dprintk(XENLOG_WARNING, "spfn %lx ~ epfn %lx pxm %x node %x"
-            "is not numa valid", spfn, epfn, pxm, node);
+        printk(XENLOG_WARNING
+               "pfn range %lx..%lx PXM %x node %x is not NUMA-valid\n",
+               spfn, epfn, pxm, node);
         return -EINVAL;
     }
 
